package org.wa5znu.rpsk;

import danger.app.*;
import danger.ui.*;
import danger.util.*;
import danger.util.DEBUG;
import danger.system.Config;

import java.io.IOException;
import java.io.InputStream;
import java.io.DataInputStream;
import java.io.OutputStream;
import java.net.Socket;
import java.net.UnknownHostException;

/**
 * Implements the main application class.
 *
 */
public class rpsk extends Application implements Resources, Commands
{
  final static int WATERFALL_WIDTH = 240;
  final static int WATERFALL_HEIGHT = 160;
  /**
   * Frequency range is 0-4000; FFT has 1024 elemements.
   * Scale is 4000/1024=3.90625.
   * We need to get this magic scale factor from the server.
   */
  final static int FREQUENCY_SCALE_MUL = 390625;
  final static int FREQUENCY_SCALE_DIV = 100000;
  final static int LINE_LENGTH = 40; // crock
  final static int SCREEN_LENGTH = 12; // CROCK
  final static int CHUNK = 32;	// crock.
  /**
   * Read this many lines from server before we resync.
   */
  final static int LINES_PER_BATCH = 16;

  /**
   * Creates the main application class.
   *
   */
  public rpsk() {
    //	Create our application's main window
    mWindow = new MainWindow();
    // Set it's title.
    mWindow.setTitle(getString(kStrRpsk));

    System.err.println("Use this to print debug messages...");
    DEBUG.p("Another way to print debug messages...");
  }


  /**
   * Handles events.  Called automatically whenever the application receives an event.
   *
   */
  public boolean receiveEvent(Event e) {
    switch (e.type) {
    case kCmdClearRX:
      mWindow.clearRX();
      break;
    case kCmdClose:
      suspending = true;
      mWindow.close();
      break;
    case kCmdOpen:
      suspending = false;
      mWindow.start();
      break;
    }
    return (super.receiveEvent(e));
  }


  //*	--------------------	launch

  /**
   * Handles the launch event.  Called once whenever the application is launched.
   *
   */
  public void launch() {
    DEBUG.p("Rpsk: launch");
  }

  //*	--------------------	resume

  /**
   * Handles the resume event.  Called automatically whenever the application is resumed.
   *
   */
  public void resume() {
    DEBUG.p("Rpsk: resume");
    suspending = false;
    mWindow.start();
  }

  //*	--------------------	Suspend

  /**
   * Handles the Suspend event.  Called automatically whenever the application is suspended.
   *
   */
  public void suspend() {
    DEBUG.p("Rpsk: suspend");
    suspending = true;
    mWindow.close();
  }

  static MainWindow mWindow;
  static boolean suspending = false;
}

