package org.wa5znu.rpsk;

import danger.app.*;
import danger.ui.*;
import danger.util.*;
import danger.util.DEBUG;
import danger.system.Config;

import java.io.IOException;
import java.io.InputStream;
import java.io.DataInputStream;
import java.io.OutputStream;
import java.net.Socket;
import java.net.UnknownHostException;

/**
 * Implements the main window class.
 *
 */
class MainWindow extends ScreenWindow implements Resources, Commands
{
  boolean working = false;

  public MainWindow() {
    mBounds = getBounds();
    mFont = Font.findBoldSystemFont();

    buildActionMenu();
  }

  final void buildActionMenu() {
    Menu menu = getActionMenu();

    menu.removeAllItems();
    menu.addFromResource(Application.getCurrentApp().getResources(), kRpskMenuResID);
  }

  public void paint(Pen inPen) {
    clear(inPen);
    if (waterfallBitmap != null) {
      /* synchronized(waterfallBitmap) */ {
	inPen.drawBitmap(0, 0, waterfallBitmap, 0, 0, rpsk.WATERFALL_WIDTH, rpsk.WATERFALL_HEIGHT);
      }
    }
    inPen.setColor(Color.BLACK);
    inPen.drawRect(mBounds);
    inPen.setFont(mFont);
    drawTextWrapped(inPen, messageBuffer);
    {
      inPen.setColor(phaseGood ? Color.RED : Color.GREEN);
      int x = frequency % rpsk.WATERFALL_WIDTH;
      inPen.setSize(3, 3);
      inPen.drawLine(x, PHASE_Y_OFFSET, x+phaseTable[phase][0], PHASE_Y_OFFSET+phaseTable[phase][1]);
      inPen.setSize(1, 1);
      inPen.drawOval(x-PHASE_Y_OFFSET+1, 1, x+PHASE_Y_OFFSET+1, PHASE_Y_OFFSET*2+1);
    }
    {
      if (frequency != requestedFrequency) {
	inPen.setColor(Color.BLUE);
	inPen.setSize(1, 1);
	inPen.drawLine(requestedFrequency, 0, requestedFrequency, rpsk.WATERFALL_HEIGHT);
	invalidate();
      }
    }
  }

  private void drawTextWrapped(Pen pen, StringBuffer messageBuffer) {
    String message = messageBuffer.toString();
    Font font = pen.getFont();
    int leading = (font.getAscent() + font.getDescent());
    int lineLength = rpsk.LINE_LENGTH;
    int lineNumber = 0;
    int end = 0;
    int len = message.length();
    while (len > 0 && end < len) {
      int start = lineNumber*lineLength;
      end = start + lineLength;
      if (end > len) 
	end = message.length();
      String lineString = message.substring(start, end);
      pen.drawText(1,(++lineNumber)*leading, lineString);
      if (lineNumber > rpsk.SCREEN_LENGTH) {
	// crock
	messageBuffer.setLength(0);
	lineNumber = 0;
      }
    }
  }

  Bitmap waterfallBitmap;

  int row = 0;
  Socket socket=null;
  DataInputStream is=null;
  OutputStream os=null;
  StringBuffer messageBuffer = new StringBuffer();
  AlertWindow alert = null;

  private void open() throws IOException {
    show();
    row = 0;
    if (socket == null || is == null) {
      setSubTitle("Opening connection...");
      System.err.println("about to make socket");
      socket = new Socket("wa5znu.org", 3125);
      // System.err.println("Socket " +socket+ " made");
      // socket.setTcpNoDelay(true);
      System.err.println("Socket TCP_NODELAY set");
      System.err.println("Remove this GNU BufferedInputStream dependency");
      is = new DataInputStream(new BufferedInputStream(socket.getInputStream()));
      os = socket.getOutputStream();
      initBitmap();
      requestSetSpectrumWidth(waterfallBitmap.getStride());
      working = true;
    }
    
  }

  boolean openSafely() {
    try {
      open();
      return true;
      // } catch (danger.io.NoServiceException e) {
      // try {Thread.sleep(1000); } catch (InterruptedException ie) { }
    } catch (UnknownHostException e) {
      if (!rpsk.suspending) { 
	showError(e); 
      } else {
	e.printStackTrace();
	DEBUG.p(e);
      }
    } catch (IOException e) {
      if (!rpsk.suspending) {
	showError(e);
	try {Thread.sleep(1000); } catch (InterruptedException ie) { }
      } else {
	e.printStackTrace();
	DEBUG.p(e);
      }
    }
    return false;
  }

  void start() {
    new Thread() {
      public void run() {
	boolean needsOpen = true;
	while (!rpsk.suspending) {
	  if (needsOpen && openSafely()) {
	    needsOpen = false;
	    try {
	      while(!rpsk.suspending) {
		if (! readLines(rpsk.LINES_PER_BATCH)) {
		  DEBUG.p("Breaking because readLines returned false");
		  DEBUG.p("rpsk: trying to recover");
		  close();
		  needsOpen = true;
		  break;
		} else {
		  DEBUG.p("Readlines returned true");
		}
	      }
	      DEBUG.p("rpsk.suspending#1="+rpsk.suspending);
	    } catch (Exception e) {
	      if (!rpsk.suspending) {
		showError(e);
		DEBUG.p("rpsk: trying to recover");
		close();
		needsOpen = true;
	      } else {
		DEBUG.p("Error during suspend -- ignoring");
		e.printStackTrace();
		DEBUG.p(e);
	      }
	    }
	  }
	}
	DEBUG.p("rpsk.suspending#2="+rpsk.suspending);	
      }
    }.start();
  }

  void showError(Exception e) {
    e.printStackTrace();
    DEBUG.p(e);
    if (alert == null) {
      alert = new AlertWindow();
      alert.setAutoDismiss(250, null);
    }
    String msg = e.getMessage();
    if (msg == null) msg = e.toString();
    if (msg == null) msg = e.getClass().getName();
    alert.setMessage("error: " + msg);
    alert.show();
  }

  void close() {
    working = false;
    alert = null;
    // waterfallBitmap = null;
    setSubTitle("Closing connection...");
    closeInputStream();
    closeOutputStream();
    closeSocket();
    clearBitmap();
  }

  void closeInputStream() {
    System.err.println("Closing inputStream");
    try {
      if (is != null) is.close(); 
    } catch (Exception e2) { 
      socket = null;
      os = null;
      System.err.println("could not close InputStream : " + e2); 
    } finally {
      is = null;
    }
    System.err.println("inputStream closed");
  }

  void closeOutputStream() {
    System.err.println("Closing outputStream");
    try {
      if (os != null) os.close(); 
    } catch (Exception e2) { 
      socket = null;
      System.err.println("could not close OutputStream : " + e2); 
    } finally {
      is = null;
    }
    System.err.println("outputstream closed");
  }

  void closeSocket() {
    System.err.println("Closing socket");
    try { 
      if (socket != null) socket.close(); 
    } catch (Exception e2) { 
      System.err.println("could not disconnect socket : " + e2);
    } finally {
      socket = null;
    }
    System.err.println("socket closed");
  }

  void clearRX() {
    messageBuffer = new StringBuffer();
    invalidate();
  }

  Bitmap initBitmap() {
    if (waterfallBitmap == null) {
      waterfallBitmap = new Bitmap(null, rpsk.WATERFALL_WIDTH, rpsk.WATERFALL_HEIGHT, ColorSpace.GRAY8, 0);
      // waterfallBitmapPen = waterfallBitmap.createPen();
    }
    return waterfallBitmap;
  }

  void clearBitmap() {
    if (waterfallBitmap != null) {
      /* synchronized(waterfallBitmap) */ {
	if (false) {
	  // This gets colorspace errors on the device
	  Pen waterfallBitmapPen = waterfallBitmap.createPen();
	  waterfallBitmapPen.setColor(Color.BLACK);
	  waterfallBitmapPen.fillRect(0, 0, waterfallBitmap.getWidth(), waterfallBitmap.getHeight());
	} else {
	  int len = waterfallBitmap.getStride() * waterfallBitmap.getHeight();
	  byte[] bits = waterfallBitmap.getPixels();
	  for (int i = 0; i < len; i++) {
	    bits[0] = 0;
	  }
	}
	invalidate();
      }
    }
  }

  boolean readLines(int nlines) throws IOException {
    if (waterfallBitmap == null) initBitmap();
    int stride = waterfallBitmap.getStride();
    byte[] bits = waterfallBitmap.getPixels();
    // Pen waterfallBitmapPen = waterfallBitmap.createPen();
    // waterfallBitmapPen.setColor(Color.RED);
    System.err.println("Starting to process " + nlines + " lines");
    for (int line = 0; working && line < nlines; line++) {
      process(is);
      int len = is.readUnsignedShort();
      if (len != rpsk.WATERFALL_WIDTH) {
	throw new RuntimeException("Asked for " + rpsk.WATERFALL_WIDTH + " but got " + len);
      }
      /* synchronized(waterfallBitmap) */ {
	row = (row+1) % rpsk.WATERFALL_HEIGHT;
	{
	  int nextRow = (row+1) % rpsk.WATERFALL_HEIGHT;
	  for (int n = 0; n < stride; n++) 
	    bits[nextRow*stride+n] = -128; // white
	}
	try {
	  int offset = 0;
	  int readLen = 0;
	  while (len > 0) {
	    int l = (len > rpsk.CHUNK) ? rpsk.CHUNK : len;
	    // System.err.println("Reading " + l + " bytes at " + row*stride+offset);
	    readLen = is.read(bits, row*stride+offset, l);
	    if (readLen < 0) {
	      throw new RuntimeException("failing: readLen is " + readLen);
	    }
	    offset += readLen;
	    len -= readLen;
	  }
	  // System.err.println("Got line " + row + " read " + readLen);
	} catch (ArrayIndexOutOfBoundsException e) {
	  DEBUG.p("row = " + row+" stride="+stride+" len="+len+" row*stride="+row*stride);
	  throw e;
	}
      }
      invalidate();
    }
    System.err.println("Read " + nlines + " lines done");
    requestSetSpectrumWidth(waterfallBitmap.getStride());
    return working;
  } 

  private void process(DataInputStream is) throws IOException {
    boolean done = false;
    while (working && !done) {
      int b = is.readUnsignedByte();
      switch(b) {
	// System.err.println("read " + ((char)b));
      case -1:
	throw new java.io.EOFException();
      case 'S':
	done = true;
	break;
      case 'C':
	{
	  int len = is.readUnsignedByte();
	  if (len == -1) throw new RuntimeException("lost reading C len");
	  for (int i = 0; working && i < len; i++) {
	    char c = (char)is.readUnsignedByte();
	    messageBuffer.append(c);
	  }
	}
	break;

      case 'B':
	{
	  int bMilliHz = (is.readUnsignedShort() * 256 + is.readUnsignedByte());
	  setSignalBandwidth(bMilliHz / 1000);
	}
	break;
      case 'F':
	setFrequency(is.readUnsignedShort());
	break;

      case 'p':
      case 'P':
	phaseGood = (b == 'P');
	phase = is.readUnsignedByte();
	invalidate();
	break;
      case 'M':
	{
	  String m = readMode(is);
	  if (m != null) {
	    // modeChoiceBox.select(m);
	    DEBUG.p("Mode = " + m);
	  }
	}
	break;
      default:
	System.err.println("Unexpected " + b + " " +  ((char)b));
	// We don't know why it gets out of sync, but the other clients
	// don't have this problem, so just work around it by giving up
	// on this connection.
	close();
	done = true;
	break;
      }
    }
  }

  /**
   * M followed by byte count follwed by mode name
   * BPSK
   * QPSK
   * MFSK16
   * MFSK8
   * PSK63
   * RTTY
   */
  private String readMode(InputStream inputStream) throws IOException {
    StringBuffer modeBuffer = new StringBuffer();
    int len = inputStream.read();
    if (len < 0) return null;
    int togo = len;

    while (togo > 0) {
      int c = inputStream.read();
      if (c < 0) {
	return null;
      }
      modeBuffer.append((char) c);
      togo -= 1;
    }
    String mode = modeBuffer.toString();
    // modeChoiceBox.select(mode);
    return mode;
  }



  int frequency = 0;
  int phase = 0;
  boolean phaseGood = false;

  // When the server tells us the frequency has changed
  void setFrequency(int f) {
    frequency = f;
    requestedFrequency = f;	// it takes so long we might as well live with it.
    System.err.println("Got frequency " + f);
    updateFrequency();
  }

  void setSignalBandwidth(int x) {
    DEBUG.p("Signal bandwidth received: " + x);
    updateFrequency();
  }

  void updateFrequency() {
    int x = (frequency*rpsk.FREQUENCY_SCALE_MUL) / rpsk.FREQUENCY_SCALE_DIV;
    if (frequency == requestedFrequency) {
      setSubTitle(x+"");
    } else {
      int r = (requestedFrequency*rpsk.FREQUENCY_SCALE_MUL) / rpsk.FREQUENCY_SCALE_DIV;
      setSubTitle(x+" "+r);
    }
  }

  // When we tell the server the frequency changed.
  // Unfortunately we don't get a response quickly enough and
  // need to maintain our own idea of sync.
  int requestedFrequency = 0;
  void requestSetFrequency(int f) {
    try {
      synchronized(os) {
	os.write('F');
	os.write(f / 256);
	os.write(f & 255);
	requestedFrequency = f;
	os.flush();
      }
      updateFrequency();
    } catch (Exception e) {
      showError(e);
    }
  }

  // When we tell the server how much spectrum we want
  void requestSetSpectrumWidth(int width) {
    // setSubTitle("Requesting spectrum " + width);
    System.err.println("Requesting spectrum " + width);
    try {
      synchronized(os) {
	os.write('S');
	os.write(width / 256);
	os.write(width & 255);
	os.flush();
      }
    } catch (Exception e) {
      showError(e);
    }
  }

  public final void renderSplashScreen(View view, Pen pen) {
    try {
      if (waterfallBitmap != null) 
	pen.drawBitmap(0, 0, waterfallBitmap);
    } catch (Exception e) {
      DEBUG.p("renderSplashScreen: " + e);
      e.printStackTrace();
    }
  }

  //*	--------------------	eventWidgetUp

  public boolean
  eventWidgetUp(int inWhichWidget, Event inEvent)
  {
    boolean	consumed = false;

    switch (inWhichWidget)
      {
      case Event.DEVICE_WHEEL:
      case Event.DEVICE_MULTIPLE_WHEEL:
      case Event. DEVICE_WHEEL_PAGE_UP:
	{
	  boolean big= (inEvent.modifierIsActive(Event.EVENT_MODIFIER_MENU_BUTTON));
	  int f = requestedFrequency -
	     (big ? rpsk.WATERFALL_WIDTH : inEvent.getWheelCount());
	  if (f < 0)
	    f = 0;	  
	  requestSetFrequency(f);
	  if (big) {
	    System.err.println("moving to " + f);
	    clearRX();
	  }
	  consumed = true;
	}
	break;
      case Event.DEVICE_WHEEL_BUTTON:
	break;
      case Event.DEVICE_ARROW_UP:
	break;
      case Event.DEVICE_ARROW_DOWN:
	break;
      case Event.DEVICE_ARROW_LEFT:
	break;
      case Event.DEVICE_ARROW_RIGHT:
	break;
      }
    return (consumed || super.eventWidgetUp(inWhichWidget, inEvent));
  }

  //*	--------------------	eventWidgetDown

  public boolean
  eventWidgetDown(int inWhichWidget, Event inEvent)
  {
    boolean	consumed = false;
    switch (inWhichWidget)
      {
      case Event.DEVICE_WHEEL:
      case Event.DEVICE_MULTIPLE_WHEEL:
      case Event. DEVICE_WHEEL_PAGE_DOWN:
	{
	  boolean big = (inEvent.modifierIsActive(Event.EVENT_MODIFIER_MENU_BUTTON));
	  int f = requestedFrequency +
	     (big ? rpsk.WATERFALL_WIDTH : inEvent.getWheelCount());
	  if (f > 1023)
	    f = 1023;
	  requestSetFrequency(f);
	  if (big) {
	    System.err.println("Moving to " + f);
	    clearRX();
	  }
	  consumed = true;
	}	
	break;
      case Event.DEVICE_WHEEL_BUTTON:
	break;
      case Event.DEVICE_ARROW_UP:
	break;
      case Event.DEVICE_ARROW_DOWN:
	break;
      case Event.DEVICE_ARROW_LEFT:
	break;
      case Event.DEVICE_ARROW_RIGHT:
	break;
      }
    return (consumed || super.eventWidgetDown(inWhichWidget, inEvent));
  }

  public boolean eventShortcut(char c, Event e) {
    switch (c) {
      // We only want Menu-Shift-T here, not Menu-t.
    case 'T': 
      toggleFullScreen();
      return true;
    default:
      break;
    }
    return super.eventShortcut(c, e);
  }

  private void toggleFullScreen() {
    setFullScreen(!isFullScreen());
    invalidate();
  }


  //*	--------------------------------	Member Variables

  Rect				mBounds;
  Font				mFont;


  final static int PHASE_Y_OFFSET = 14;
  final static byte[][]phaseTable = {
    { 0, -14 },// 0
    { 0, -14 },// 1
    { 1, -14 },// 2
    { 1, -14 },// 3
    { 1, -14 },// 4
    { 2, -14 },// 5
    { 2, -14 },// 6
    { 2, -14 },// 7
    { 3, -14 },// 8
    { 3, -14 },// 9
    { 3, -14 },// 10
    { 4, -14 },// 11
    { 4, -14 },// 12
    { 4, -13 },// 13
    { 5, -13 },// 14
    { 5, -13 },// 15
    { 5, -13 },// 16
    { 6, -13 },// 17
    { 6, -13 },// 18
    { 6, -13 },// 19
    { 7, -12 },// 20
    { 7, -12 },// 21
    { 7, -12 },// 22
    { 8, -12 },// 23
    { 8, -12 },// 24
    { 8, -12 },// 25
    { 8, -11 },// 26
    { 9, -11 },// 27
    { 9, -11 },// 28
    { 9, -11 },// 29
    { 9, -10 },// 30
    { 10, -10 },// 31
    { 10, -10 },// 32
    { 10, -10 },// 33
    { 10, -9 },// 34
    { 11, -9 },// 35
    { 11, -9 },// 36
    { 11, -9 },// 37
    { 11, -8 },// 38
    { 12, -8 },// 39
    { 12, -8 },// 40
    { 12, -8 },// 41
    { 12, -7 },// 42
    { 12, -7 },// 43
    { 12, -7 },// 44
    { 13, -6 },// 45
    { 13, -6 },// 46
    { 13, -6 },// 47
    { 13, -5 },// 48
    { 13, -5 },// 49
    { 13, -5 },// 50
    { 13, -4 },// 51
    { 14, -4 },// 52
    { 14, -4 },// 53
    { 14, -3 },// 54
    { 14, -3 },// 55
    { 14, -3 },// 56
    { 14, -2 },// 57
    { 14, -2 },// 58
    { 14, -2 },// 59
    { 14, -1 },// 60
    { 14, -1 },// 61
    { 14, -1 },// 62
    { 14, 0 },// 63
    { 14, 0 },// 64
    { 14, 0 },// 65
    { 14, 1 },// 66
    { 14, 1 },// 67
    { 14, 1 },// 68
    { 14, 2 },// 69
    { 14, 2 },// 70
    { 14, 2 },// 71
    { 14, 3 },// 72
    { 14, 3 },// 73
    { 14, 3 },// 74
    { 14, 4 },// 75
    { 14, 4 },// 76
    { 13, 4 },// 77
    { 13, 5 },// 78
    { 13, 5 },// 79
    { 13, 5 },// 80
    { 13, 6 },// 81
    { 13, 6 },// 82
    { 13, 6 },// 83
    { 12, 7 },// 84
    { 12, 7 },// 85
    { 12, 7 },// 86
    { 12, 8 },// 87
    { 12, 8 },// 88
    { 12, 8 },// 89
    { 11, 8 },// 90
    { 11, 9 },// 91
    { 11, 9 },// 92
    { 11, 9 },// 93
    { 10, 9 },// 94
    { 10, 10 },// 95
    { 10, 10 },// 96
    { 10, 10 },// 97
    { 9, 10 },// 98
    { 9, 11 },// 99
    { 9, 11 },// 100
    { 9, 11 },// 101
    { 8, 11 },// 102
    { 8, 12 },// 103
    { 8, 12 },// 104
    { 8, 12 },// 105
    { 7, 12 },// 106
    { 7, 12 },// 107
    { 7, 12 },// 108
    { 6, 13 },// 109
    { 6, 13 },// 110
    { 6, 13 },// 111
    { 5, 13 },// 112
    { 5, 13 },// 113
    { 5, 13 },// 114
    { 4, 13 },// 115
    { 4, 14 },// 116
    { 4, 14 },// 117
    { 3, 14 },// 118
    { 3, 14 },// 119
    { 3, 14 },// 120
    { 2, 14 },// 121
    { 2, 14 },// 122
    { 2, 14 },// 123
    { 1, 14 },// 124
    { 1, 14 },// 125
    { 1, 14 },// 126
    { 0, 14 },// 127
    { 0, 14 },// 128
    { 0, 14 },// 129
    { -1, 14 },// 130
    { -1, 14 },// 131
    { -1, 14 },// 132
    { -2, 14 },// 133
    { -2, 14 },// 134
    { -2, 14 },// 135
    { -3, 14 },// 136
    { -3, 14 },// 137
    { -3, 14 },// 138
    { -4, 14 },// 139
    { -4, 14 },// 140
    { -4, 13 },// 141
    { -5, 13 },// 142
    { -5, 13 },// 143
    { -5, 13 },// 144
    { -6, 13 },// 145
    { -6, 13 },// 146
    { -6, 13 },// 147
    { -7, 12 },// 148
    { -7, 12 },// 149
    { -7, 12 },// 150
    { -8, 12 },// 151
    { -8, 12 },// 152
    { -8, 12 },// 153
    { -8, 11 },// 154
    { -9, 11 },// 155
    { -9, 11 },// 156
    { -9, 11 },// 157
    { -9, 10 },// 158
    { -10, 10 },// 159
    { -10, 10 },// 160
    { -10, 10 },// 161
    { -10, 9 },// 162
    { -11, 9 },// 163
    { -11, 9 },// 164
    { -11, 9 },// 165
    { -11, 8 },// 166
    { -12, 8 },// 167
    { -12, 8 },// 168
    { -12, 8 },// 169
    { -12, 7 },// 170
    { -12, 7 },// 171
    { -12, 7 },// 172
    { -13, 6 },// 173
    { -13, 6 },// 174
    { -13, 6 },// 175
    { -13, 5 },// 176
    { -13, 5 },// 177
    { -13, 5 },// 178
    { -13, 4 },// 179
    { -14, 4 },// 180
    { -14, 4 },// 181
    { -14, 3 },// 182
    { -14, 3 },// 183
    { -14, 3 },// 184
    { -14, 2 },// 185
    { -14, 2 },// 186
    { -14, 2 },// 187
    { -14, 1 },// 188
    { -14, 1 },// 189
    { -14, 1 },// 190
    { -14, 0 },// 191
    { -14, 0 },// 192
    { -14, 0 },// 193
    { -14, -1 },// 194
    { -14, -1 },// 195
    { -14, -1 },// 196
    { -14, -2 },// 197
    { -14, -2 },// 198
    { -14, -2 },// 199
    { -14, -3 },// 200
    { -14, -3 },// 201
    { -14, -3 },// 202
    { -14, -4 },// 203
    { -14, -4 },// 204
    { -13, -4 },// 205
    { -13, -5 },// 206
    { -13, -5 },// 207
    { -13, -5 },// 208
    { -13, -6 },// 209
    { -13, -6 },// 210
    { -13, -6 },// 211
    { -12, -7 },// 212
    { -12, -7 },// 213
    { -12, -7 },// 214
    { -12, -8 },// 215
    { -12, -8 },// 216
    { -12, -8 },// 217
    { -11, -8 },// 218
    { -11, -9 },// 219
    { -11, -9 },// 220
    { -11, -9 },// 221
    { -10, -9 },// 222
    { -10, -10 },// 223
    { -10, -10 },// 224
    { -10, -10 },// 225
    { -9, -10 },// 226
    { -9, -11 },// 227
    { -9, -11 },// 228
    { -9, -11 },// 229
    { -8, -11 },// 230
    { -8, -12 },// 231
    { -8, -12 },// 232
    { -8, -12 },// 233
    { -7, -12 },// 234
    { -7, -12 },// 235
    { -7, -12 },// 236
    { -6, -13 },// 237
    { -6, -13 },// 238
    { -6, -13 },// 239
    { -5, -13 },// 240
    { -5, -13 },// 241
    { -5, -13 },// 242
    { -4, -13 },// 243
    { -4, -14 },// 244
    { -4, -14 },// 245
    { -3, -14 },// 246
    { -3, -14 },// 247
    { -3, -14 },// 248
    { -2, -14 },// 249
    { -2, -14 },// 250
    { -2, -14 },// 251
    { -1, -14 },// 252
    { -1, -14 },// 253
    { -1, -14 },// 254
    { 0, -14 },// 255
  };

}
